/*
** Module   : FrameGuiImpl.java
** Abstract : Frame implementation for GUI
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20141010 Created initial version.
** 002 HC  20141021 Implemented window move/position support.
** 003 EVL 20141127 Inserting this container into mouse source chain.
** 004 CA  20141205 Avoid unnecessary conversion, as the physical location/dimension can be 
**                  determined from existing APIs.  Replaced saved dimension with real dimension
**                  on resize.
** 005 EVL 20150311 Adding default button processing for GUI.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.Font;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;//
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.ScrollPaneGuiImpl.ShowBars;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.Widget;

/**
 * A <code>Frame</code> is a top-level window with a title and a border. 
 */
@SuppressWarnings("unchecked")
public class FrameGuiImpl
extends Frame<GuiOutputManager>
{
   /** Screen driver */
   private GuiDriver gd;

   /** Output manager */
   private GuiOutputManager screen;
   
   /** Coordinate converter */
   private CoordinatesConversion cc;

   /** Color resolver */
   private GuiColorResolver gc;

   /** Font resolver */
   private GuiFontResolver gf;

   /** Frame scroll pane */
   private ScrollPaneGuiImpl fs;
      
   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public FrameGuiImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, (FrameConfig) cfg);
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // set frame default colors
      FrameConfig config = (FrameConfig) cfg;
      gc = new GuiColorResolver(gd, config, SysColor.COLOR_WINDOW, SysColor.COLOR_WINDOWTEXT);
      // box and title colors
      gc.boxColor     = gc.sysFgColor;
      gc.titleFgColor = gc.sysBgColor;
      gc.titleBgColor = gd.getSysColor(SysColor.COLOR_ACTIVECAPTION);
      // output manager
      screen = screen();
      cc = screen.coordinates();
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, WidgetConfig wcfg)
   {
      super.initialize(id, wcfg);
      gf = new GuiFontResolver((Window<?>) ancestor(), gd, config());
      configScrollPane();
   }

   /**
    * Draw the peer container. The methods calls the draw() method for
    * ScrollPane instance for the peer container. In addition the required
    * clipping rectangle is calculated for ChildFrame containers because they
    * must be rendered inside the parent container.
    * <p>
    * Drawing for the redirected terminal differs. Labels/headers are drawn
    * only once per page, instead of the usual once per frame.
    * 
    * @param   origin
    *          The coordinate of the upper-left corner for peer container in
    *          character units.
    * @param   size
    *          The size of the peer container in character units.
    */
   @Override
   public void draw(Point origin, Dimension size)
   {
      if (!isVisible())
      {
         return;         
      }
      
      // determine frame colors
      gc.colors();
   
      // resolve font
      final FontCache cached = gf.font();
            
      final NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();

      final NativeRectangle clip = clipRectangle(p, d);

      // resize scroll pane
      resizeScrollPane(hasTitle() ? cached.fontHeight : 0);

      gd.draw(p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            // draw frame
            gd.setColor(gc.bgColor);
            gd.fillRect(0, 0, d.width, d.height);
            
            if (frameScroll != null)
            {
               if (!down)
               {
                  frameScroll.draw();
               }
            }
            
            // draw frame title
            drawCaption(config().title);

            // draw fame box if defined
            if (config().box)
            {
               gd.setColor(gc.boxColor);
               gd.drawRect(0, 0, d.width - 1, d.height - 1);               
            }
         }

         /**
          * Draw frame title.
          * 
          * @param   title
          *          Frame title.
          */
         private void drawCaption(String title)
         {
            if (hasTitle())
            {
               // draw caption rectangle
               gd.setColor(gc.titleBgColor);
               gd.fillRect(0, 0, d.width, cached.fontHeight);

               // calculate text position
               int textWidth = gd.getTextWidth(title, cached.font);
               int x = (d.width - textWidth) / 2;
               int y = cached.font.getSize();

               // draw title string
               gd.setColor(gc.titleFgColor);
               gd.setGuiFont(cached.font);
               gd.setFontStyle(Font.BOLD);
               gd.drawString(title, x, y);
               
               if (config().box)
               {
                  // draw caption box 
                  gd.setColor(gc.boxColor);
                  gd.drawRect(0, 0, d.width - 1, cached.fontHeight);
               }
            }
         }
      });      
   }
   
   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units).
    * Each container should search for own widgets. Because widgets are relative positioned 
    * inside containers the container physical position is adjusted before search.  
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or {@code null} if none found.
    */
   @Override
   public Widget<GuiOutputManager> findMouseSource(NativePoint p)
   {
      // the point incoming is related to frame origin, we need to substract the frame title
      NativePoint pCheck = hasTitle() ? p.minus(fs.physicalLocation()) : p;
      
      return fs.findMouseSource(pCheck);
   }

   /**
    * Reset colors.
    * 
    * @param   config
    *          Widget configuration.
    */
   protected void resetColors(BaseConfig config)
   {
      // TODO should be tested
      config.bgcolor = -1;
      config.fgcolor = -1;
   }
   
   /**
    * Processing for default button in aq frame.
    *
    * @return <code>TRUE</code> if event be re-routed to default button,
    *         <code>FALSE</code> otherwise.
    */
   @Override
   public boolean processDefaultButton(KeyInput event)
   {
      // scan the event source
      Widget src = event.source();
      // check if the source is the field-level widget but not button
      if (src instanceof FillIn            ||
          src instanceof SelectionList     ||
          src instanceof SelectionListBody ||
          src instanceof RadioSet          ||
          src instanceof ComboBox          ||
          src instanceof ToggleBox         ||
          src instanceof Editor)
      {
         // calculate new event target
         Button<GuiOutputManager> bTarget =
            (Button<GuiOutputManager>) OutputManager.resolveWidget(config().defaultButtonId);
         // button default attribute must also be set to make this work
         if (bTarget != null && bTarget.config().isDefault)
         {
            // route enter key to default button
            bTarget.processKeyEvent(event);
            return true;
         }
      }
      // not field level widget
      return false;
   }
   
   /**
    * Compare colors.  This method must be implemented in subclasses and provide
    * specific implementation as follow:
    * <p>
    * ChUI implementation must compare dcolor and pfcolor.
    * <p>
    * GUI implementation must compare bgcolor and fgcolor.
    * 
    * @param   config
    *          Widget configuration.
    * @param   colors
    *          Colors to be compared.
    *          
    * @return  The colors structure after comparison.         
    */
   @Override
   protected Colors compareColors(BaseConfig config, Colors colors)
   {
      colors.bgEquals = colors.bgcolor == config.bgcolor;
      colors.fgEquals = colors.fgcolor == config.fgcolor;
      return colors;
   }
   
   /**
    * Configure frame scroll pane
    */
   @SuppressWarnings("rawtypes")
   private void configScrollPane()
   {
      fs = (ScrollPaneGuiImpl) this.frameScroll;
      fs.setShowBars(ShowBars.ALWAYS);
      fs.config().sysbgcolor = SysColor.COLOR_WINDOW;
      fs.setScrollBars(ScrollBar.Bars.BOTH);
      // register scroll pane and its widgets for mouse events
      WidgetRegistry<?> reg = OutputManager.instance().getRegistry();
      reg.addWidget((Widget) fs);
      fs.registerWidgets((WidgetRegistry<GuiOutputManager>) reg);
   }
      
   /**
    * Resize frame scroll pane.
    * 
    * @param   captionHeight
    *          The height of title bar.
    */
   private void resizeScrollPane(int captionHeight)
   {
      // location
      NativePoint np = new NativePoint(0, captionHeight);
      fs.setPhysicalLocation(np.x, np.y);
      fs.setLocation(cc.pointFromNative(np));
      // size
      NativeDimension nd = physicalDimension();
      nd.width  = nd.width;
      nd.height = nd.height - captionHeight;
      fs.setSize(cc.dimensionFromNative(nd));
   }
}
